

// define pins to segments -----------------------------------------------
#define Gs  (1<<RB0)
#define Cs  (1<<RB1)
#define Hs  (1<<RB2)
#define Ds  (1<<RB3)
#define Es  (1<<RB4)
#define As  (1<<RB5)
#define Fs  (1<<RB6)
#define Bs  (1<<RB7)


//----- define signes ------------------------
#define _D1  (Bs|Cs)
#define _D2  (As|Bs|Ds|Es|Gs)
#define _D3  (As|Bs|Cs|Ds|Gs)
#define _D4  (Bs|Cs|Fs|Gs)
#define _D5  (As|Cs|Ds|Fs|Gs)
#define _D6  (As|Cs|Ds|Es|Fs|Gs)
#define _D7  (As|Bs|Cs)
#define _D8  (As|Bs|Cs|Ds|Es|Fs|Gs)
#define _D9  (As|Bs|Cs|Ds|Fs|Gs)
#define _D0  (As|Bs|Cs|Ds|Es|Fs)

#define _U  (Bs|Cs|Ds|Es|Fs)
#define _u  (Cs|Ds|Es)
#define _P  (As|Bs|Es|Fs|Gs)
#define _Top As
#define _Bottom Ds
#define _Middle Gs
#define _Space  0
#define _S  (As|Cs|Ds|Fs|Gs)
#define _O  (As|Bs|Cs|Ds|Es|Fs)
#define _o  (Cs|Ds|Es|Gs)
#define _C  (As|Ds|Es|Fs)
#define _c  (Ds|Es|Gs)
#define _A  (As|Bs|Cs|Es|Fs|Gs)
#define _L  (Ds|Es|Fs)
#define _I  (Bs|Cs)
#define _F  (As|Es|Fs|Gs)
#define _t  (Ds|Es|Fs|Gs)
#define _n  (Cs|Es|Gs)
#define _E  (As|Ds|Es|Fs|Gs)
#define _r  (Es|Gs)
#define _H  (Bs|Cs|Es|Fs|Gs)
#define _i  Cs
#define _b  (Cs|Ds|Es|Fs|Gs)
#define _d  (Bs|Cs|Ds|Es|Gs)

// define PORTC pins
   //control common pins of indicator
#define portc_sign0_pin  RC7
#define portc_sign1_pin  RC4
#define portc_sign2_pin  RC5
#define portc_sign3_pin  RC6
   //mesuring control
#define portc_U1_pin  1<<RC0
#define portc_U2_pin  1<<RC3
#define portc_U3_pin  1<<RC1

   //measuring current on/off
#define U1_on  TRISC &= ~portc_U1_pin
#define U2_on  TRISC &= ~portc_U2_pin
#define U3_on  TRISC &= ~portc_U3_pin
#define U1_off  TRISC |= portc_U1_pin
#define U2_off  TRISC |= portc_U2_pin
#define U3_off  TRISC |= portc_U3_pin

// Define PORTA pins
    // ADC inputs
#define A_direct RA0  //
#define A_OPAMP RA1   //
#define A_ind  RA3    //   for ind. testing and help for batt. measure
#define A_bat  RA2    //   for batt. measure
    // mosfet control
#define porta_Mos_pin  RA5
    // charger control
#define A_charger  RA4  // DO NOT CHANGE !!!

//Mosfet on/off
#define Mos_on  PORTA |= (1<<porta_Mos_pin)
#define Mos_off  PORTA &= ~(1<<porta_Mos_pin)


// The work option
#define Time_to_sleep  60 // time to sleep (sec.)
#define Time_to_calibrate  30 // time to calibrate  (sec.)



// Define signes for common cathode
const unsigned char Gen_[] = {_D0,_D1,_D2,_D3,_D4,_D5,_D6,_D7,_D8,_D9};
                               // define digits

unsigned char  Res_10[8] = {0,0,0,0,0,0,0,0}, Cap_10[8] = {0,0,0,0,0,0,0,0};
           // signs to OUT, R, C. Elements 0 - 3 is
           // directly going to ind., 4 - 7 is next
unsigned int  Res_, Cap_, U_bat;
unsigned short volatile Timer=0xFF, TimeCount=0xFF, TimeSec=0;
char Cali, Res0;//calibration coeff. for U1 (100+-50); the probes own resistance

unsigned short  Flags = 0;    // Flags bits means:
#define f_anode  Flags.F0 //0 - 1-common anode, 0-cathode
#define f_calib  Flags.F1 //1 - 1- count the time to calibration
#define f_sleep  Flags.F2 //2 - 1- count the time to sleep
#define f_charge  Flags.F3 //3 - 1-batt. is charging
#define f_low    Flags.F4 //4 - 1- the battery voltage is low
#define f_lowest Flags.F5 //5 - 1- the battery voltage is low
#define f_full   Flags.F6 //6 - 1- the battery is full charged
#define f_indbat Flags.F7 //7 - 1 - now indicated battery voltage







//********************* Procedure of pause 1 - 127 ms (by 1ms step) ***********
// In parameter - Delay (ms)
// Needs global variable Timer which increments in 0.5 ms interrups

void pause_ms(unsigned char Delay ) {
  char Delay1;
  Delay1 = Delay;
  //Delay = Delay << 1; // intervals 1 ms (*2)
  Delay += Timer;
  while (Timer != Delay); // wait (Delay) ms or (in-parameter) ms
  asm clrwdt;
  Delay1 += Timer;
  while (Timer != Delay1); // wait (Delay) ms or (in-parameter) ms
  asm clrwdt;
 }

//******************************************************************************





//================= Pause x sec ================================================
// In parameter - time for pause (sec)

void pause_s(char Sec){
   char i;
   Sec <<= 2;   // *4
   for (i=0; i<Sec; i++) pause_ms(125);
   }
//------------------------------------------------------------------------------







// ********** SYNCRONIZATION WITH TMR1 INTERRUPS *******************************
// *****************************************************************************
void int_sync(void){
  PIE1.TMR1IE = 0;
  while (PIR1.TMR1IF==0);
  PIE1.TMR1IE = 1;
  asm clrwdt;
  asm{
   nop;
   nop;
   }
 }      // we have more than 450us to next interrupt

//*****************************************************************************








 //************************* PWM off *******************************************

 void PWM1_off(void) {
 CCP1CON = CCP1CON&0xF0;
 T2CON &=(~(1<<TMR2ON));
 TRISC |= (1<<RC2);     //rc2 to in
 TRISC &= (~(1<<RC2));  //rc2 to out
 PORTC &= (~(1<<RC2));  // rc2 to 0
 }
 //*****************************************************************************




  //************************* PWM on *******************************************

 void PWM1_on(void) {
 T2CON |=(1<<TMR2ON);
 CCP1CON = CCP1CON|0x0C;
 TRISC &= (~(1<<RC2));
 }
 //*****************************************************************************







//********* Set 8-bit PWM to meandr and switch on*******************************

void PWM1_Initial(void)  {
              //Init. ------------
  T2CON = 0X05; // PRESCALER TO 4
  PR2 = 0xFF;   // f = 4 kHz

  CCPR1L = 0;
  CCP1CON =0;

                 // Set duty for meandr
  CCPR1L = 0x7F;
 }

//*****************************************************************************









//********** BIN to BCD (0 - 9999)  PROCEDURE ***********************************
// In parameter - 1. 0 - 9999 int,  
//                2. pointer to 4-byte BCD array (simbols to segments)
// writes BCD digits to pointed array
//******************************************************************************
void int_to_ind(unsigned int In_int, char *In_10) {
  unsigned char s, Sign[4];

  s = (unsigned char)(In_int / 1000);  // thousands
  if (s>9) return ;
  Sign[3] = Gen_[s];
  In_int = In_int % 1000;

  s = (unsigned char)(In_int  / 100);  // handreds
  Sign[2] = Gen_[s];
  s = (unsigned char)(In_int % 100);
  Sign[1] = Gen_[(unsigned char)(s / 10)];   // decades
  Sign[0] = Gen_[(unsigned char)(s % 10)];  // units
  
  int_sync();
  In_10 += 4;
  for (s=0; s<=3; s++) {   // save to current array
    *(In_10+s) = Sign[s];
   }
 }
 //*****************************************************************************





 // ------------ Discharging the capacitor at 125 ms ---------------------------
 void discharge(void) {
   Mos_on;  // discharging on
   pause_ms(125);
   Mos_off;  // discharging off

 }
// -----------------------------------------------------------------------------






//====================== current non-linearity correction ======================
// added to result of ADC coefficient for linearity

unsigned int corr_ADC(unsigned int inADC) {
 char in_4, i;
 const char corr[] = {50, 80, 100, 120, 140, 152, 165, 179, 190, 200, 210, 220, 230, 240, 249, 255};
 in_4 = (char)(inADC >> 2);
 for (i=0; i<=15; i++) {
   if (corr[i] >= in_4) break;
   }
 return inADC + i;
 }
//==============================================================================







// ================== The first test of connected element ======================
// Return: 0 - nothing connected, 1 - capasitor/resistor, 2 - ionistor
// Running time 0.25 - 0.5 s

char in_test(void) {
 char i;

 ADCON1 = 0x80;  // RIGHT JAST.;  PORTA TO ANALOG
 ADCON0 = (A_direct<<3)| 0b10000001; // Fosc/32, A_direct, ON

 for (i=0; i<=3; i++){   // up to 4 cicles (<=500ms)
  discharge();
  pause_ms(1);   // 1ms to rise the voltage when not connect
  ADCON0.GO = 1;          // start
  while (ADCON0.GO);      // wait to convertions end

  if (ADRESL <= 15)  return 1; // capasitor or resistor  if <=50mV
  }
  
  Mos_on;  // open the mosfet
  pause_ms(1);
  ADCON0.GO = 1;          // start
  while (ADCON0.GO);      // wait to convertions end
  Mos_off;  // switch off mosfet
  
  if (ADRESL <= 3)  return 0; // not connected
     else return 2;       // ionistor


}

//------------------------------------------------------------------------------




 
 
 
 
 
//============== Measuring of capacity (prototype) =============================

void C_measure(char Ux, struct Meter *S);

// -----------------------------------------------------------------------------









// ============= Measuring of R or ESR   prototype =============================

void R_measure(char Ux, struct Meter *S);

// -----------------------------------------------------------------------------










// ====================== ESR + C or R measuring ===============================
// return flag means: 
// 0, 1 bits:   0 - U3, 1 - U2, 2 - U1,
// bits 3-7 to set: 3 - ionistor, 4 - resistor, 5 - capasitor, 6 - C > 999 uF,
//                  7 - out of range
//
// writes to global variables Res_ and Cap_

char measure(void) {
 struct Meter{unsigned int V1, V2, V3, Vc, Counter;} M ; // V1-V3 - start,
       // first, second voltages (measure ESR); Vc - max voltage of
       // capacitor, Counter - number of 26 us intervals (measuring C);
 unsigned long Cap;
 char flag=0, x;
 ADCON1 = 0x80;  // RIGHT JAST.;  PORTA TO ANALOG
 ADCON0 = (A_OPAMP<<3)| 0b10000001; // Fosc/32, A_OPAMP, ON
 Delay_us(50);
 
       // measure by U1
 flag = 2;                        // try U1
 x =  portc_U1_pin;
 R_measure(portc_U1_pin, &M); //

 
 if (M.V3 > 800) { // ------------ADC too big decrease the current and retry
 
     flag = 1;                        // try U2
     x =  portc_U2_pin;
     R_measure(portc_U2_pin, &M); //

     if (M.V3 > 800) { // ------------ADC too big decrease the current and retry
     
          flag = 0;                   // try U3
          x =  portc_U3_pin;
          R_measure(portc_U3_pin, &M);//

          if (M.V3 > 800) return 0x80;  // exit with 'out of range'
         } // ---------------------------------------------------------------

     } // -------------------------------------------------------------------


 C_measure(x, &M);        // measure the time to charge to 0.1V
 
 M.Vc = corr_ADC(M.vc);   //
 M.V1 = corr_ADC(M.v1);   //
 M.V2 = corr_ADC(M.v2);   //
 M.V3 = corr_ADC(M.v3);   // current non-linearity correction
 
 
 
 // ------------------------ R, ESR measuring -----------------
   // Ur = 2*U2 - U3 - U1
   // Resr = 330*Nadc/(1024*Kop) , Kop - OPAMP gain
   // Kop = 32.23   =>  Resr = Nadc (x0.01 Ohm)
                                   
   // Res_ = 2*M.v2 - M.v3 - M.v1;

 Res_ = 2*M.v2 - M.v3 - M.v1;
 Res_ = (unsigned long)Res_*100/Cali;  // correction for calibr.
 
 switch (Flag){          // calculate the current coeff. for Res0
   case 0: x = 100; break;
   case 1: x = 10; break;
   default: x = 1;   //
   }

 Res_ -= Res0/x;  // Corr. for probes own resis.

 if (M.Vc < M.V1 + 10) {  // test for resistor  (if Vc-V2 < 10)
    Flag.F4 = 1;
    return Flag;    // return -> resistor
    }


 // ------------------------ C measuring -----------------
 // C = (I*T)/U;     for Kopamp = 32,23 ==>
 // C(x0.1uF) = 285*Count*X/(Vc-V1);  X = 1, 10 or 100 for Flag = 0,1 or 2
 //        (28,5us - real measured pulse width )
 switch (Flag){          // calculate the current coeff.
   case 0: x = 1; break;
   case 1: x = 10; break;
   default: x = Cali;   // for Flag=2 calibration coeff. insted 100
   }
 
 Cap = 285 * (unsigned long)M.Counter;
 Cap = Cap * x / (M.Vc - M.V1);
 
            // tests for elements and save to int --------------
 if (Cap < 100000){
    Flag.F5 = 1;   // Capacitor
    if (Cap > 9999){
       Flag.F6 = 1;   // >999uF
       Cap_ = (unsigned int)(Cap / 10);
      }
      else Cap_ = (unsigned int)Cap;
   }
   else {
      Flag.F3 = 1;   //  ionistor
      Cap_ = (unsigned int)(Cap / 10000);
      }
           // --------------------------------------------------


 return flag;
 }
















// ============= Measuring of R or ESR =========================================
// changed the v1 - v2 of structure
// in : current type (v1 - v3) , point to structure Meter

void R_measure(char Ux, struct Meter *S) {
 unsigned int V1, V2, V3;
 discharge();  //--> 26us after end of interrups
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 V1 = (ADRESH << 8) + ADRESL;       // save the start voltage

   // start charging by Ux
 TRISC &= ~Ux;    // Ux on
          //  for interval 26us for waiting same next measuring
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 V2 = (ADRESH << 8) + ADRESL;       // save the voltage
 asm nop;
          // measure v2
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 V2 = (ADRESH << 8) + ADRESL;       // save the voltage
 asm {
   nop;   // add for 26us
   nop;
   nop;}
         // measure v3
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 V3 = (ADRESH << 8) + ADRESL;       // save the start voltage
 asm nop;
 TRISC |= Ux;  //switch off current (current was switched on within 26*3 =78us)

 S->V1 = V1;
 S->V2 = V2;
 S->V3 = V3;     // save to structure

 }


//==============================================================================









//============== Measuring of capacity =========================================

void C_measure(char Ux, struct Meter *S){
 unsigned int Vc0, Vc1, Counter;
 Vc0 = S->V3 - S->V2;                     //
 if (Vc0 > 1023) Vc0 = 0; // if <0        //
 Vc0 = Vc0 * 2 + S->V1;                   // Vc0 = v3-vr = 2*(v3-v2)+v1
               // after V3 capacitor already charged within 26us
               // therefor began with Counter=2
     // In next cycle the capacitor charging by 26us pulses until voltage is not
     // out of range or total about to 100ms at maximum (Counter = 4000)
 for (Counter=2; Counter <= 4000; Counter++) {
   if (TMR1L > 230) int_sync(); // if <50us to end interrupt interval then sync.
  
   ADCON0.GO = 1; // measuring voltage after previous charging
   asm nop;
   TRISC &= ~Ux;    // Ux on , charge
   while (ADCON0.GO);      //
   Vc1 = (ADRESH << 8) + ADRESL;
   if (Vc1 > 800) break;   // Exit if top limit
   TRISC |= Ux;  //Ux off , not charge
   Vc0 = Vc1;
   Delay_us(15);   // delay for OPAmp
   }                      // One step of cycle about 48us

 TRISC |= Ux;  //Ux off after break

 S->Vc = Vc0;
 S->Counter = Counter;

 }

// -----------------------------------------------------------------------------










 // ================ Charge indication =========================================
 
void ind_charge(void) {
 char Bat;
 
 Bat = U_bat>>2;

 if (Bat<85) {  // u < 3.4v
    Cap_10[7]=_C;
    Cap_10[6]=_Space;
    Cap_10[5]=_Space;
    Cap_10[4]=_Space;
    Res_10[7]=_C | _Middle;
    Res_10[6]=_Space;
    Res_10[5]=_Space;
    Res_10[4]=_Space;
    return;
   }
 
 if (Bat<91) {  // u < 3.64v
    Cap_10[7]=_C;
    Cap_10[6]=_Top | _Bottom;
    Cap_10[5]=_Space;
    Cap_10[4]=_Space;
    Res_10[7]=_C | _Middle;
    Res_10[6]=_Top | _Bottom | _Middle;
    Res_10[5]=_Space;
    Res_10[4]=_Space;
    return;
   }
 if (Bat<97) {  // u < 3.88v
    Cap_10[7]=_C;
    Cap_10[6]=_Top | _Bottom;
    Cap_10[5]=_Top | _Bottom;
    Cap_10[4]=_Space;
    Res_10[7]=_C | _Middle;
    Res_10[6]=_Top | _Bottom | _Middle;
    Res_10[5]=_Top | _Bottom | _Middle;
    Res_10[4]=_Space;
    return;
   }
 if (Bat<103) {  // u < 4.12v
    Cap_10[7]=_C;
    Cap_10[6]=_Top | _Bottom;
    Cap_10[5]=_Top | _Bottom;
    Cap_10[4]=_Top | _Bottom;
    Res_10[7]=_C | _Middle;
    Res_10[6]=_Top | _Bottom | _Middle;
    Res_10[5]=_Top | _Bottom | _Middle;
    Res_10[4]=_Top | _Bottom | _Middle;
    return;
   }
 Cap_10[7]=_C;
 Cap_10[6]=_Top | _Bottom;
 Cap_10[5]=_Top | _Bottom;
 Cap_10[4]=_Top | _Bottom | _D1;
 Res_10[7]=_C | _Middle;
 Res_10[6]=_Top | _Bottom | _Middle;
 Res_10[5]=_Top | _Bottom | _Middle;
 Res_10[4]=_D3;

 }

//------------------------------------------------------------------------------













//=================== Indication 'out of range' ================================
void w_out(void) {
       Cap_10[7]=_c;
       Cap_10[6]=_Space;
       Cap_10[5]=_Bottom;
       Cap_10[4]=_Bottom;          // 'c __'

       Res_10[7]=_r;
       Res_10[6]= Gen_[9];
       Res_10[5]= Gen_[9];      //     _
       Res_10[4]=_Top;          // 'r99 '  - limitless
   }
//==============================================================================



// ================== Indication of charged ionistor ===========================
void w_ionistor(void){
       Cap_10[7]=_i;
       Cap_10[6]=_o;
       Cap_10[5]=_n;
       Cap_10[4]=_Bottom;          // 'ion_'

       Res_10[7]=_U;
       Res_10[6]=_Top;
       Res_10[5]=_Middle;          //   _
       Res_10[4]=_Bottom;          // 'U -_'  - DISCHARGE
  }
// =============================================================================








// ================= Indication of 'calib. error' ==============================
 void w_error(void) {
       Cap_10[7]=_C;
       Cap_10[6]=_A;
       Cap_10[5]=_L;
       Cap_10[4]=_I;          // 'CALI'

       Res_10[7]=_E;
       Res_10[6]=_r;
       Res_10[5]=_r;
       Res_10[4]=_Bottom;          // 'Err_'
 pause_s(3);
 }
//------------------------------------------------------------------------------












// ================= Indication of 'end calib.' ================================
 void w_end(void) {
       Cap_10[7]=_C;
       Cap_10[6]=_A;
       Cap_10[5]=_L;
       Cap_10[4]=_I;          // 'CALI'

       Res_10[7]=_E;
       Res_10[6]=_n;
       Res_10[5]=_d;
       Res_10[4]=_Bottom;          // 'End_'
 pause_s(2);
 }
//------------------------------------------------------------------------------









//  =============== Indication of 'not connect' or batt. voltage ===============
//  Indication of 'not connect' but indicate batt. voltage each 16s within 2s

 void w_notconnect(){
  static char Flag = 0;

  if (f_indbat==1) {
   if (TimeSec.F1==1) Flag = 2;
   if (Flag==0) {  // do next block one time
     Flag = 1;
     int_to_ind(U_bat, &Res_10);
     Res_10[3+4] = _u;
     Res_10[2+4] |= Hs; // +comma = uX.XX
     if (f_low==1) {
        Cap_10[3+4] = _b;
        Cap_10[2+4] = _Bottom;
        Cap_10[1+4] = _L;
        Cap_10[0+4] = _o;
       }
       else {
          Cap_10[3+4] = _b;
          Cap_10[2+4] = _A;
          Cap_10[1+4] = _t;
          Cap_10[0+4] = _t;
          }
     return;
     }
   }
     else Flag = 0;
     
  if (Flag == 1) return;
 
  if (f_charge==1){ // indication if charged
    ind_charge();
    return;
    }
 
  Cap_10[7]=_Space;
  Cap_10[6]=_Middle;
  Cap_10[5]=_Space;
  Cap_10[4]=_Space;          // ' -  '
  Res_10[7]=_Space;
  Res_10[6]=_Space;
  Res_10[5]=_Middle;
  Res_10[4]=_Space;          // '  - '
  }
//==============================================================================






//====================== turn off the ind. =====================================
void w_space(void) {
       Cap_10[7]=_Space;
       Cap_10[6]=_Space;
       Cap_10[5]=_Space;
       Cap_10[4]=_Space;

       Res_10[7]=_Space;
       Res_10[6]=_Space;
       Res_10[5]=_Space;
       Res_10[4]=_Space;
    }
//------------------------------------------------------------------------------








// ======================= Calibration =========================================
char calibration(void) {
 unsigned int x;
 char i;
 
       Cap_10[7]=_C;
       Cap_10[6]=_A;
       Cap_10[5]=_L;
       Cap_10[4]=_I;          // 'CALI'

       Res_10[7]=_C;
       Res_10[6]=_A;
       Res_10[5]=_L;
       Res_10[4]=_I;    // 'CALI'
 pause_s(2);

 ADCON1 = 0x80;  // RIGHT JAST.;  PORTA TO ANALOG
 ADCON0 = (A_OPAMP<<3)| 0b10000001; // Fosc/32, A_OPAMP, ON
 Delay_us(50);

       Cap_10[7]=_O;
       Cap_10[6]=_P;
       Cap_10[5]=_E;
       Cap_10[4]=_n;          // 'OPEn'

       Res_10[7]=_P;
       Res_10[6]=_r;
       Res_10[5]=_o;
       Res_10[4]=_b;          // 'Prob'     open the probs
       
 U2_on;
 for (i=0; i<100; i++) {
    ADCON0.GO = 1;          // start
    while (ADCON0.GO);      // wait to convertions end
    if ((ADRESH == 0x03)&&(ADRESL == 0xFF)) break;  // exit if probes not conn.
    pause_ms(100);
    }
 U2_off;
 if (i==50) return 0;    // return if probes connected >10s after mesage (error)
 pause_s(1);
 
       // Calibration U1 current ---------------
 Mos_on; //
 U1_on;  //   measure for 1 Ohm
 pause_ms(5);
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 Mos_off; //
 U1_off;
 x =  (ADRESH << 8) + ADRESL;
 if (x>400) return 0;   // R>4  - error
 Cali = (char)(x * 97 / 200);  // -3%  (R1 and Rds of MOSFET)
 EEPROM_Write(1, Cali); // save the current calib.

 w_space();  // blank the ind
 pause_s(2);  // 3 sec. ind. mesage
          //   end U1 calibr. ------------------
 
 
          // Probe res. calibration --------------
       Cap_10[7]=_C;
       Cap_10[6]=_L;
       Cap_10[5]=_O;
       Cap_10[4]=_S;          // 'CLOS'

       Res_10[7]=_P;
       Res_10[6]=_r;
       Res_10[5]=_o;
       Res_10[4]=_b;          // 'Prob'      CLOSE probes

 for (i=0; i<100; i++) {
    ADCON0.GO = 1;          // start
    while (ADCON0.GO);      // wait to convertions end
    if (ADRESH == 0) break;  // exit if probes is connect
    pause_ms(100);
    }
 if (i==100) {
    Res0 = 0;
    EEPROM_Write(2, 0); // save 0 to the probes res.
    return 1;   // return if probes not connected >10s after mesage with Res0 = 0
    }
    
 pause_s(1);
 U1_on;  //   measure for probes res.
 pause_ms(1);
 ADCON0.GO = 1;          // start
 while (ADCON0.GO);      // wait to convertions end
 U1_off;
 if (ADRESH) return 0;   // R>255  - error
 Res0 = (unsigned int)ADRESL*100/Cali;
 EEPROM_Write(2, Res0); // save the probes res.

 w_space();    // blank the ind/
 pause_s(3);  // 3 sec. ind. mesage

 }

 // ----------------------------------------------------------------------------











//===================== To  SLEEP ==============================================
void to_sleep(void) {


  while (1) {
     ADCON1 = 0x07; // porta to digit
     ADCON0 = 0;    // adc off
     PORTB = 0xFF;          // IND.OFF BY TURNING TO HI SEGMENTS AND ANODE/CATH.
     PORTC |= (1<<portc_sign0_pin)|(1<<portc_sign1_pin)|(1<<portc_sign2_pin)|(1<<portc_sign3_pin);
     PORTA = (1<<A_charger)|(1<<A_ind); // charger pin + ind pin to 1. anothers to 0
     PWM1_off();
     PIE1.TMR1IE = 0; // disable int.
     asm sleep;
     
     ADCON1 = 0x80;  // RIGHT JAST.;  PORTA TO ANALOG
     ADCON0 = (A_direct<<3)| 0b10000001; // Fosc/32, A_direct, ON
     Delay_us(100);
     ADCON0.GO = 1;          // start
     while (ADCON0.GO);      // wait to convertions end
     if ((ADRESL<30) || (ADRESL>150)) break;  //exit if closed probes or plug in
                                              // the charged capacitor

     PORTA = 0; // charge on, prepare measure
     ADCON1 = 0;  // LEFT JAST.;  PORTA TO ANALOG
     ADCON0 = (A_bat<<3)| 0b10000001; // Fosc/32, A_direct, ON
     Delay_us(100);
     ADCON0.GO = 1;          // start
     PORTA = (1<<A_charger)|(1<<A_ind); // charge off,
     while (ADCON0.GO);      // wait to convertions end
     
     if (ADRESH>155) {  //exit if charger is connect
        TimeSec = 8;          // prepare battery test
        f_indbat = 0;         //
        f_sleep = 0;   //reset sleep flag
        PWM1_on();
        PIE1.TMR1IE = 1;
        return;
        }
   }

  PORTA = 1<<A_charger ; // only charger pin to 1. anothers to 0
  f_sleep = 0;   //reset sleep flag
  TimeSec = 0;   // and sec. timer
  PIE1.TMR1IE = 1;
  PWM1_on();
 }


// -------------------------------------------------------------------------------











// ========================== Measuring the battery voltage ====================
void batt_test(void) {
 char x;
 f_indbat = 1;
 
 x = PORTA & (1<<A_charger); // 0- charging
 
 ADCON1 = 0;  // LEFT JAST.; PORTA TO ANALOG
 ADCON0 = (A_bat<<3)| 0b10000001; // Fosc/32, A_bat, ON
 Delay_us(50);
 
 if (!x) {   //  measure charger voltage -----------
    ADCON0.GO = 1;          // start
    while (ADCON0.GO);      // wait to convertions end
   }
   else {
      PORTA &= ~(1<<A_charger); // reset pin A_charger
      pause_ms(10);
      ADCON0.GO = 1;          // start
      PORTA |= 1<<A_charger; // SET pin A_charger
      while (ADCON0.GO);      // wait to convertions end
     }                    //-----------------------
 if ((ADRESH < 155)||(ADRESH > 233)) f_charge = 0; //if <4.5v or >5.5v dont charge
      else f_charge = 1; //if between 4.5v and 5.5v  => charge
 
 ADCON1 = 0x80;  // RIGHT JAST.;  PORTA TO ANALOG
 pause_ms(10);
 if (x) {   //  measure battery voltage -----------
    ADCON0.GO = 1;          // start
    while (ADCON0.GO);      // wait to convertions end
   }
   else {
      PORTA |= 1<<A_charger; // charge off
      pause_ms(10);
      ADCON0.GO = 1;          // start
      PORTA &= ~(1<<A_charger); // charge on
      while (ADCON0.GO);      // wait to convertions end
     }                    //-----------------------
 U_bat = (ADRESH<<8)+ADRESL;
 U_bat = (long)U_bat * 330 / 1024 + 250;

            // Battery service -------------
 if (U_bat > 415) f_full = 1; // >4.15v => full
 if (U_bat < 405) f_full = 0; // <4.05v => not full
 if((f_charge==1)&&(f_full==0)) {
    PORTA &= ~(1<<A_charger); // charge on
    }
    else {
       PORTA |= 1<<A_charger; // charge off
       }
           // Flags of batt. setting -------------
 if (U_bat<345) f_low = 1; // <3.45v => low batt.
    else f_low = 0;
 if (U_bat<330) f_lowest = 1; // <3.3v => very low
    else f_lowest = 0;
 


 }

//==============================================================================




















// ==================== THE MAIN PROCEDURE =====================================
void main() {
  char mflag, x;  // mflag - measuring result flag

  OPTION_REG = 0b00001011;// pullup on, prescaler -8 to WDT. (wdt period ~0.15s)

  TRISC = ~((1<<portc_sign0_pin)|(1<<portc_sign1_pin)|(1<<portc_sign2_pin)|(1<<portc_sign3_pin));
            // common and mosfet control to out
            

  // ------------ test for common anode or cathode -----------------------------
  PORTC = 0;
  ADCON1 = 4;  // LEFT JAST.; PINS 0,1,3 PORTA TO ANALOG
  ADCON0 = (A_ind<<3)| 0b10000001; // Fosc/32, A_ind, ON
  Delay_us(50);
  ADCON0.GO = 1;          // start
  while (ADCON0.GO);      // wait to convertions end
  
  f_anode = 0;                    // cathode
  if (ADRESH > 200)  f_anode = 1; // anode
  // ---------------------------------------------------------------------------
  
  if (f_anode==0) PORTB = 0;
     else PORTB = 1;     // ind. off
  PORTC |= ((portc_U1_pin)|(portc_U2_pin)|(portc_U3_pin)); // U1, U2, u3 pins to HI
  
  if (EEPROM_Read(0)!= 0) { // Inic. or read EEPROM-------------------
     EEPROM_Write(0,0);
     EEPROM_Write(1,100);  // calibration coeff. (100+-50)
     EEPROM_Write(2,0);    // the probes  resistance =0
     Cali = 100;
     Res0 = 0;
    }
    else {
       Cali = EEPROM_Read(1);
       Res0 = EEPROM_Read(2);
      }        // ----------------------------------------------------
      
  
  TRISB = 0;
  PORTA = 1<<A_charger ; // only charger pin to 1. anothers to 0
  TRISA = ~((1<<porta_Mos_pin)|(1<<A_charger)|(1<<A_ind)); // Mos pin,
                                     // charge pin and ind. pin to out
  OPTION_REG.F7; // PULL UP OFF
  
  T1CON = 0x31;     // prescaler -8, int. clock, TMR1 ON
  INTCON = 0B11000000; // ON,  PEIE ON
  PIE1.TMR1IE = 1;   // TMR1 INT. ON    //period 512 us (USED ONLY TMR1L)
  
  
  // Set power of OPAMP ---------
  PWM1_Initial();
  PWM1_on();

  while(1){       // the bace cycle ---------------------------------------------------------
  
    if (TimeSec.F3==1) {
         if (f_indbat==0){
            f_indbat==1;
            batt_test();    // 1 time each 16s
          }
       }
       else f_indbat = 0;
       
                        // ------------- to sleep --------------
    if (f_charge == 0){ // if not charged
      if (f_lowest==1) to_sleep(); // too low batt. => sleep
      if (f_sleep==1) { //sleep flag is set within 1 min => sleep (or 10s if batt. is low)
        if ((TimeSec > Time_to_sleep) || ((TimeSec > 10)&&(f_low==1))) {
           to_sleep();
           }
        }
      }
                        // ------ to calibration ---------------
    if ((f_calib==1)&&(TimeSec >= Time_to_calibrate)) {//calib. flag is set within 20s => calib.
        f_calib=0;
        if (!calibration()) w_error();
            else w_end();
        TimeSec = 0;
       }


    mflag = in_test(); // first test of the connected element
    
    if (! mflag) {    // if =0  - not connected
       w_notconnect();
       f_calib = 0;
       if (f_sleep==0){   // if  lost connection  in just moment
         f_sleep = 1;     // set sleep flag
         TimeSec = 0;     // and reset seconds timer
         }
       continue;
       }
       else f_sleep = 0;

    if (mflag == 2) {    // if =3  - discharging of ionistor
       w_ionistor();
       continue;
       }

                 // C, R or ESR measuring area ---------------------------------
    mflag = measure();
    
    if (mflag.F7 == 1) {    // out of range
       w_out();
       continue;
       }
                        // Resistant to ind. ------------------
    x = mflag & 0x03;

    if (Res_ > 999) {
       if (Res_ < 2048) {
          if (x==0) Res_ = 999;        // max
              else {Res_ /= 10; x--;} // change limit for 3-digit result
          }
          else  Res_ = 0;    // if 2*v2-v1-v3 < 0 => to 0
     }

    int_to_ind(Res_, &Res_10);
    Res_10[3+4] = _r;    //  1-st is 0. Replace with 'r'
    
    if ((Res_10[2+4] == _O)&&(x!=2)){        //blank not significant zero
       Res_10[2+4] = _Space;
       if ((Res_10[1+4] == _O)&&(x==0)) Res_10[1+4] = _Space;
       }                                            //--
       
    if (x != 0) Res_10[x +4] |= Hs ;    // set comma
    
    
                        // Capacitance to ind
    if (mflag.F4 == 1) {    // resistor
      Cap_10[7]=_Bottom;
      Cap_10[6]=_r;
      Cap_10[5]=_E;
      Cap_10[4]=_S;          // '_rES'
      
      if (f_calib==0){    // if the probes is connected now directly or resistor
         f_calib = 1;     // set calibration flag
         TimeSec = 0;     // and reset seconds timer
         }
      continue;
      }
      else f_calib = 0;

    if (mflag.F5 == 1) {    // Capasitor
      int_to_ind(Cap_, &Cap_10);
      if (Cap_10[3+4] == _O) {
        Cap_10[3+4] = _c;    // c   if 1-st is 0
        if (Cap_10[2+4] == _O) Cap_10[2+4] = _Space; // blank unsigificant zero
        }
      if (mflag.F6 == 0) Cap_10[1 +4] |= Hs ;    // set comma
      }
      
    if (mflag.F3 == 1) {    // ionistor
      if (Cap_ < 1000) {
        int_to_ind(Cap_, &Cap_10);
        }
        else {
          Cap_10[6]=Gen_[9];
          Cap_10[5]=Gen_[9];        //   _
          Cap_10[4]=_Top;          // '99 ' if >999
          }
      Cap_10[3+4] = _F;       //
      Cap_10[3+4] |= Hs;    // 'F.xxx'
      }
    
   } //------------- End of base cycle --------------------------------------------------------

  
 }

//==============================================================================










 //**************************** INTERRUPTS **************************************
void interrupt(void) {

 unsigned char To_portc=0, i;
 static unsigned char Segment = 1, SignNo = 0;         // column of indicator
 static unsigned char *Digit; // BCD to indicator pointer (Res_10 or Cap_10)

 if (PIR1.TMR1IF == 1){
    TMR1H = 0xFF;    // period 255
    Timer +=1 ;                          // 0.5 ms counter (8 bit)
    if (Timer==0){
       TimeCount +=1;                 // 128 ms counter  (8 bit), total - 32s
       if ((TimeCount & 0x07) == 0) TimeSec +=1; // Seconds counter
       }

        // Data update block ----------------------
    if ((Timer == 0)&&((0x03 & TimeCount) == 0)){
       if (TimeCount.F2 == 0) Digit = &Res_10;
          else Digit = &Cap_10;

       *(Digit) =  *(Digit+4);
       *(Digit+1) =  *(Digit+5);      // show within 0.5 s
       *(Digit+2) =  *(Digit+6);      //  for both numbers
       *(Digit+3) =  *(Digit+7);

      } // -------------------------------------------


    if (Segment.F7 == 0) Segment <<= 1;
       else {Segment = 1;
           SignNo++;
           if (SignNo >3) SignNo = 0;
           }                         // rotate left for each 4 sign


    To_portc = (1<<portc_sign0_pin)|(1<<portc_sign1_pin)|(1<<portc_sign2_pin)|(1<<portc_sign3_pin);
    

    if (f_anode==0){   // for common cathode ----------------------------------

       To_portc = PORTC | To_portc;    // PORTC >> CATHODE pins to 1, others not changed

       if(SignNo == 0) To_portc &= (~(1 << portc_sign0_pin));
       if(SignNo == 1) To_portc &= (~(1 << portc_sign1_pin));
       if(SignNo == 2) To_portc &= (~(1 << portc_sign2_pin));
       if(SignNo == 3) To_portc &= (~(1 << portc_sign3_pin));
                           // clear current cathode pin
       PORTB = 0;       // IND. off

       PORTC = To_portc |(portc_U1_pin)|(portc_U2_pin)|(portc_U3_pin);     // to portc

       PORTB = (*(Digit+SignNo)) & Segment; // set to on current segment
       }
       else {   // for common anode --------------------------------------------

          To_portc = PORTC & (~To_portc); // PORTC >> ANODE pins to 0, others not changed

          if(SignNo == 0) To_portc |= (1 << portc_sign0_pin);
          if(SignNo == 1) To_portc |= (1 << portc_sign1_pin);
          if(SignNo == 2) To_portc |= (1 << portc_sign2_pin);
          if(SignNo == 3) To_portc |= (1 << portc_sign3_pin);
                           // set current anode pin
          PORTB = 0xFF;       // IND. off

          PORTC = To_portc |(portc_U1_pin)|(portc_U2_pin)|(portc_U3_pin);    // to portc

          PORTB = ~((*(Digit+SignNo)) & Segment); // set to on current segment

          }  //-----------------------------------------------------------------
    

    PIR1.TMR1IF=0;      // clear flag TMR0IF
   
   }


 }
//******************** END OF INTERRUPS ***************************************